/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* subroutines dealing with bonded interactions
   search_bond_base - searches the data base and sets up initial vectors
                                      fbond - gets the force of the bond
   getvbond - gets the potential energy of the bond
   getvbond_intra - gets the potential energy of the bond for AA bonds within CG sites
   bonded   - check to see if to atoms are bonded
   fcbond   - used in NMA (normal mode analysis)
   fcbondn  - used in NMA (normal mode analysis)
*/

#include "md.h"

static int ibon,jbon,itbon;
static double eqbon,fkbon,bmbon;
#define RCGBOND 3.71   /* 3.71 have to make this an input variable */
#define _RCGBOND sqrt(10.)   /* 3.71 have to make this an input variable */

#define _NORESCALING_OFF
#define __TESTHANDLE

/* #define PRINT_BOND */
/*-----------------------------------------------------------------------*/
int bond_pair(int iatom,int jatom,BONDS *bonds)
{
  int i,j,itemp,nb;

  if(iatom>jatom){ itemp = iatom;iatom = jatom; jatom = itemp;}

  for(nb=0;nb<bonds->nbonds;nb++){
    i = bonds->ibond[nb]; j = bonds->jbond[nb];
    if(i>j){itemp = i; i = j; j = itemp; }
    if(iatom == i && jatom == j) return 1;
  }
  return 0;
}
    
/*-----------------------------------------------------------------------*/
void exclude_bonds(int *nexclude,int **exclude,BONDS *bonds)
{
  int i,j,nb;

  for(nb=0;nb<bonds->nbonds;nb++){
    if(bonds->ibond[nb]<bonds->jbond[nb]){
      i=bonds->ibond[nb];j=bonds->jbond[nb];
    } else {
      j=bonds->ibond[nb];i=bonds->jbond[nb];
    }
    insert(exclude[i],&nexclude[i],j);
  }
}

/*-----------------------------------------------------------------------*/
/*- add bond to selection if both ends belong an active AA/CG region ----*/
int select_bonds(int *iaacg,BONDS *bonds,int iflg)
{
  int nsel,nsel2, ib, i,j,*isel;
  int *ibond, *jbond;
  
  nsel = 0;
  
  switch(iflg){	
  case 1: /* for CG: selects all but AA-AA */
    ibond = bonds->ibond;
    jbond = bonds->jbond;
    isel = bonds->iselect;
    for(ib=0;ib<bonds->nbonds;ib++){ /* select: at least one atom is not inactive */
      if( iaacg[ibond[ib]]+iaacg[jbond[ib]] != 0 ){
        isel[nsel++]=ib;
      } 
    }
    bonds->nselect = nsel;
    break;
  case 0: /* for AA: selects AA-AA, HR-HR and AA-HR */
    ibond = bonds->ibond;
    jbond = bonds->jbond;
    isel = bonds->iselect;
    for(ib=0;ib<bonds->nbonds;ib++){ /* select: neither atom is inactive */
      if( (iaacg[ibond[ib]] !=0 ) && (iaacg[jbond[ib]] != 0) ){
        isel[nsel++]=ib;
      } 
    }
    bonds->nselect = nsel;
    break;
  case 2: /* SO3 dynamics selection: all but AA-AA */
    ibond = bonds->ibond2;
    jbond = bonds->jbond2;
    isel = bonds->isel_so3cgr;
    nsel2 = bonds->nbonds2;
    for(ib=0;ib<bonds->nbonds2;ib++){ /* CG=0 HR=1 AA=2 */
      i = iaacg[ibond[ib]];
      j = iaacg[jbond[ib]];
      if( ( i == 0 ) && ( j == 0) ){
        isel[nsel++]=ib;
      }else if( i + j != 4 ){
        isel[--nsel2]=ib;
      } 
    }
    bonds->nsel_so3cgr = nsel;
    bonds->nsel_so3hr = bonds->nbonds2 - nsel2;
    bonds->isel_so3hr = isel + nsel2;
    nsel += bonds->nsel_so3hr;
    break;
  }
  
  return nsel;
}

/*-----------------------------------------------------------------------*/
void fbond(SIMPARMS *simparms,COORDS *coords)
{
  int i,j,idx,nb,nbonds,*ibond,*jbond,*idxbond,*itypbond;
  long ibegin,iend;
  double dx,dy,dz,r,fr,dr,mb;
  double *px,*py,*pz;
  double *fx,*fy,*fz;
  double *eqbond,*fkbond,*bmorse;
	
  px = coords->px;   py = coords->py;   pz = coords->pz;
  fx = coords->fxa;  fy = coords->fya;  fz = coords->fza;
  nbonds = coords->bonds.nbonds;
  ibond = coords->bonds.ibond;
  jbond = coords->bonds.jbond;
  idxbond = coords->bonds.idxbond;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;
	
  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);
	
  fr = 0.;
  for(nb=ibegin;nb<iend;nb++){
    i = ibond[nb]; j = jbond[nb]; idx = idxbond[nb];
    dx = px[i] - px[j];  
    dy = py[i] - py[j];  
    dz = pz[i] - pz[j];
    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r-eqbond[idx];
    
    switch(itypbond[idx]){/* fr is 1/r (dv/dr) */
    case 0:
      fr = fkbond[idx]*dr/r;
      break;
    case 1:
      mb = exp(-bmorse[idx]*dr);
      fr = (2.*fkbond[idx]*bmorse[idx]*mb*(1.-mb))/r;
      break;
    }
        
    dx *= fr; dy *= fr; dz *= fr;
    fx[i] -= dx;  fy[i] -= dy;  fz[i] -= dz;
    fx[j] += dx;  fy[j] += dy;  fz[j] += dz;
  }
}
/*-------------------------------------------------------------------*/

void fbondhybrid_OLD(SIMPARMS *simparms,COORDS *coords)
{
  int i,j,idx,ib,nb,nbonds,*ibond,*jbond,*idxbond,*itypbond,*iselect,*icg,lhybrid;
  long ibegin,iend;
  double dx,dy,dz,r,fr,dr,mb,fhr;
  double *px,*py,*pz,*faacg;
  double *fx,*fy,*fz;
  double *eqbond,*fkbond,*bmorse;

  px = coords->px;   py = coords->py;   pz = coords->pz;
  fx = coords->fxa;  fy = coords->fya;  fz = coords->fza;
  nbonds = coords->bonds.nselect;
  iselect = coords->bonds.iselect;
  ibond = coords->bonds.ibond;
  jbond = coords->bonds.jbond;
  idxbond = coords->bonds.idxbond;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;
  faacg = simparms->faacg + (1-coords->lcg)*simparms->ncg;
  icg = simparms->icgunit;
  lhybrid=(simparms->icalc_type == 5);

  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);

  fr = 0.;
  for(nb=ibegin;nb<iend;nb++){
    ib = iselect[nb];
    i = ibond[ib]; j = jbond[ib]; idx = idxbond[ib];
    dx = px[i] - px[j];  
    dy = py[i] - py[j];
    dz = pz[i] - pz[j];

    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r-eqbond[idx];
    
    /* healing region switch function added here BERND */
    if(lhybrid){
      if(coords->lcg){    /* CG-CG bonds are always scaled */
	fhr = MAX(faacg[i],faacg[j]);
      }else{              /* only AA-AA bonds over different CG atoms are scaled */
	if(icg[i]==icg[j]){
	  fhr = 1.;
	}else{
	  fhr = MIN(faacg[i],faacg[j]);
	}
      }
    }else{
      fhr = 1.;
    }
    
    switch(itypbond[idx]){/* fr is 1/r (dv/dr) */
    case 0:
      fr = fkbond[idx]*dr/r;
      break;
    case 1:
      mb = exp(-bmorse[idx]*dr);
      fr = (2.*fkbond[idx]*bmorse[idx]*mb*(1.-mb))/r;
      break;
    }
    
    fr *= fhr;
    dx *= fr; dy *= fr; dz *= fr;
    fx[i] -= dx;  fy[i] -= dy;  fz[i] -= dz;
    fx[j] += dx;  fy[j] += dy;  fz[j] += dz;
  }
}
/*-----------------------------------------------------------------------*/
/* bonds in healing region                                               */
void fbond_so3(SIMPARMS *simparms,COORDS *coords,SO3 *so3,int iregion)
// it seems that only need "tor" from the so3 struct...perhaps parameter
{
  int i,j,ii,idx,ib,nb,nbonds,*ibond,*jbond,*idxbond,*itypbond,*iselect;
  int icg, jcg, *icgunit, lcoupling;
  long ibegin,iend;
  double dx,dy,dz,r,rcg,fr,dr,mb,fhr,*faacg,*ds,*vds,pot;
  double *px,*py,*pz,*pcgx,*pcgy,*pcgz, dcgx, dcgy, dcgz;
  double px1, py1, pz1, px2, py2, pz2;
  double tx1, ty1, tz1, tx2, ty2, tz2;
  double *fx,*fy,*fz,*fcgx,*fcgy,*fcgz;
  double *eqbond,*fkbond,*bmorse, (*tor)[3];

  lcoupling = (simparms->iso3dyn == 1);

  px = coords->px;   py = coords->py;   pz = coords->pz;
  pcgx = coords->next->px;   pcgy = coords->next->py;   pcgz = coords->next->pz;
  fx = coords->fxt;  fy = coords->fyt;  fz = coords->fzt;
  fcgx = coords->next->fxt;   fcgy = coords->next->fyt;   fcgz = coords->next->fzt;

  if( iregion == 0){
    /* coarse grain region selection */
    nbonds = coords->bonds.nsel_so3cgr;
    iselect = coords->bonds.isel_so3cgr;
  }else{
    /* healing region selection */
    nbonds = coords->bonds.nsel_so3hr;
    iselect = coords->bonds.isel_so3hr;		
  }
  ibond = coords->bonds.ibond2;
  jbond = coords->bonds.jbond2;
  idxbond = coords->bonds.idxbond2;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;
  icgunit = simparms->icgunit;
  faacg = simparms->faacg + simparms->ncg;
  ds = simparms->dfaacg + simparms->ncg;
  vds = simparms->vds_intra + simparms->ncg; 
  tor = so3->tor;
  
  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);
  
  fr = 0.;
  for(nb=ibegin;nb<iend;nb++){ 
    ib = iselect[nb];
    i = ibond[ib]; j = jbond[ib]; idx = idxbond[ib];

    px1 = px[i]; py1 = py[i]; pz1 = pz[i]; 
    px2 = px[j]; py2 = py[j]; pz2 = pz[j];

    icg = icgunit[i]; jcg = icgunit[j];
    
    /* renormalize for distance between CG beads */
    dcgx = pcgx[icg] - pcgx[jcg];  
    dcgy = pcgy[icg] - pcgy[jcg];
    dcgz = pcgz[icg] - pcgz[jcg];		
    rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);

    dr = RCGBOND/rcg;

    dx = dcgx * dr + px1 - px2; 
    dy = dcgy * dr + py1 - py2; 
    dz = dcgz * dr + pz1 - pz2; 
    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r-eqbond[idx];
    
    switch(itypbond[idx]){/* fr is 1/r (dv/dr) */
    case 0:
      fr = fkbond[idx]*dr/r;
      pot = 0.5*fkbond[idx]*dr*dr;
      break;
    case 1:
      mb = exp(-bmorse[idx]*dr);
      pot = fkbond[idx]*(1-mb)*(1-mb);
      fr = (2.*fkbond[idx]*bmorse[idx]*mb*(1.-mb))/r;
      break;
    }
        
    ii = (faacg[i]<faacg[j])? i:j;
    fhr = 1. - faacg[ii]; 
    vds[ii] -= pot*ds[ii];
    
    fr *= fhr;
    dx *= fr; dy *= fr; dz *= fr;
    fx[i] -= dx;  fy[i] -= dy;  fz[i] -= dz;
    fx[j] += dx;  fy[j] += dy;  fz[j] += dz;

    /* now compute coupling component on the CG blobs */
    tx1 = pz1*dy - py1*dz;   ty1 = px1*dz - pz1*dx;   tz1 = py1*dx - px1*dy;
    tx2 = py2*dz - pz2*dy;   ty2 = pz2*dx - px2*dz;   tz2 = px2*dy - py2*dx;

    tor[icg][0]+=tx1;    tor[icg][1]+=ty1;    tor[icg][2]+=tz1;
    tor[jcg][0]+=tx2;    tor[jcg][1]+=ty2;    tor[jcg][2]+=tz2;

    /* == COMcounterforce plan IN HEALING AND ATOMISTIC REGION == */
    /* an if-statement for every atom in every bond?? should be improved */
    if( simparms->iaacg[icg] < 2){
      fcgx[icg] += dx;  fcgy[icg] += dy;  fcgz[icg] += dz;
    }
    if( simparms->iaacg[jcg] < 2){
      fcgx[jcg] -= dx;  fcgy[jcg] -= dy;  fcgz[jcg] -= dz;
    }
    /* ============================================ */

    if(lcoupling){
      tx1+=tx2;  ty1+=ty2;  tz1+=tz2;
      r = 1./rcg/rcg;
      dx = (dcgy*tz1 - dcgz*ty1) * r;
      dy = (dcgz*tx1 - dcgx*tz1) * r;
      dz = (dcgx*ty1 - dcgy*tx1) * r;
      
      fcgx[icg] += dx;  fcgy[icg] += dy;  fcgz[icg] += dz;
      fcgx[jcg] -= dx;  fcgy[jcg] -= dy;  fcgz[jcg] -= dz;
    }
  }
}
/*-----------------------------------------------------------------------*/
/* force from so3 bonds in healing region for CG blob icg                */
void fbond_so3hr_CGR2HR(SIMPARMS *simparms,COORDS *coords,int icgnow)
{
  int i,j,idx,ib,nb,nbonds,*ibond,*jbond,*idxbond,*itypbond,*iselect;
  int icg, jcg, *icgunit, inow, jnow, jcgnow, *iaacg;
  long ibegin,iend;
  double dx,dy,dz,r,rcg,fr,dr,mb,amass;
  double *px,*py,*pz,*pcgx,*pcgy,*pcgz, dcgx, dcgy, dcgz;
  double px1, py1, pz1, px2, py2, pz2;
  double *fx,*fy,*fz,*fcgx,*fcgy,*fcgz;
  double *eqbond,*fkbond,*bmorse;


  px = coords->px;   py = coords->py;   pz = coords->pz;
  pcgx = coords->next->px;   pcgy = coords->next->py;   pcgz = coords->next->pz;
  fx = coords->fxa;  fy = coords->fya;  fz = coords->fza;
  fcgx = coords->next->fxa;   fcgy = coords->next->fya;   fcgz = coords->next->fza;

  /* healing region selection */
  nbonds = coords->bonds.nsel_so3hr;
  iselect = coords->bonds.isel_so3hr;	

  ibond = coords->bonds.ibond2;
  jbond = coords->bonds.jbond2;
  idxbond = coords->bonds.idxbond2;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;
  icgunit = simparms->icgunit;
  iaacg = simparms->iaacg;
  
  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);
  
  fr = 0.;
  for(nb=ibegin;nb<iend;nb++){
    ib = iselect[nb];
    i = ibond[ib]; j = jbond[ib];
    icg = icgunit[i]; jcg = icgunit[j];
    if(icg==icgnow){
      inow = i;
      jnow = j;
      jcgnow = jcg;
    }else{
      if(jcg!=icgnow) continue;
      inow = j;
      jnow = i;
      jcgnow = icg;
    }

    px1 = px[inow]-pcgx[icgnow];  py1 = py[inow]-pcgy[icgnow];  pz1 = pz[inow]-pcgz[icgnow];
    if(iaacg[jcgnow]==2){
      px2 = px[jnow];  py2 = py[jnow];  pz2 = pz[jnow];
    }else{
      px2 = px[jnow]-pcgx[jcgnow];  py2 = py[jnow]-pcgy[jcgnow];  pz2 = pz[jnow]-pcgz[jcgnow];
    }
    idx = idxbond[ib];
    
    /* renormalize for distance between CG beads */
    dcgx = pcgx[icgnow] - pcgx[jcgnow];  
    dcgy = pcgy[icgnow] - pcgy[jcgnow];
    dcgz = pcgz[icgnow] - pcgz[jcgnow];		
    rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);

    dr = RCGBOND/rcg;

    dx = dcgx * dr + px1 - px2; 
    dy = dcgy * dr + py1 - py2; 
    dz = dcgz * dr + pz1 - pz2; 
    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r-eqbond[idx];
    
    switch(itypbond[idx]){/* fr is 1/r (dv/dr) */
    case 0:
      fr = fkbond[idx]*dr/r;
      break;
    case 1:
      mb = exp(-bmorse[idx]*dr);
      fr = (2.*fkbond[idx]*bmorse[idx]*mb*(1.-mb))/r;
      break;
    }

    dx *= fr; dy *= fr; dz *= fr;
    fx[inow] -= dx;  fy[inow] -= dy;  fz[inow] -= dz;

    /* == COMcounterforce == */
    amass = 1./coords->next->amass[icgnow];
    fcgx[icgnow] += dx*amass;  fcgy[icgnow] += dy*amass;  fcgz[icgnow] += dz*amass;
  } 
}
/*-----------------------------------------------------------------------*/
/* force from so3 bonds in healing region for CG blob icg                */
void fbond_so3hr_HR2CGR(SIMPARMS *simparms,COORDS *coords,SO3 *so3,int icgnow)
{
  int i,j,idx,ib,nb,nbonds,*ibond,*jbond,*idxbond,*itypbond,*iselect;
  int icg, jcg, *icgunit, inow, jnow, jcgnow, *iaacg,iregion;
  long ibegin,iend;
  double dx,dy,dz,r,rcg,fr,dr,mb,amass;
  double *px,*py,*pz,*pcgx,*pcgy,*pcgz, dcgx, dcgy, dcgz;
  double px1, py1, pz1, px2, py2, pz2;
  double *fx,*fy,*fz,*fcgx,*fcgy,*fcgz;
  double *eqbond,*fkbond,*bmorse;

    px = coords->px;   py = coords->py;   pz = coords->pz;
  pcgx = coords->next->px;   pcgy = coords->next->py;   pcgz = coords->next->pz;
  fx = coords->fxa;  fy = coords->fya;  fz = coords->fza;
  fcgx = coords->next->fxa;   fcgy = coords->next->fya;   fcgz = coords->next->fza;

  ibond = coords->bonds.ibond2;
  jbond = coords->bonds.jbond2;
  idxbond = coords->bonds.idxbond2;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;
  icgunit = simparms->icgunit;
  iaacg = simparms->iaacg;
    
  fr = 0.;
  /* healing region selection plus CG region selection (unfortunately both are needed) */
  nbonds = coords->bonds.nsel_so3cgr;
  iselect = coords->bonds.isel_so3cgr;
  for(iregion=0;iregion<2;++iregion){
  
  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);
  for(nb=ibegin;nb<iend;nb++){
    ib = iselect[nb];
    i = ibond[ib]; j = jbond[ib];
    icg = icgunit[i]; jcg = icgunit[j];
    if(icg==icgnow){
      inow = i;
      jnow = j;
      jcgnow = jcg;
    }else{
      if(jcg!=icgnow) continue;
      inow = j;
      jnow = i;
      jcgnow = icg;
    }

    px1 = px[inow];  py1 = py[inow];  pz1 = pz[inow];
    if(iaacg[jcgnow]==2){
      px2 = px[jnow];  py2 = py[jnow];  pz2 = pz[jnow];
    }else{
      px2 = px[jnow]-pcgx[jcgnow];  py2 = py[jnow]-pcgy[jcgnow];  pz2 = pz[jnow]-pcgz[jcgnow];
    }
    idx = idxbond[ib];
    
    /* renormalize for distance between CG beads */
    dcgx = pcgx[icgnow] - pcgx[jcgnow];  
    dcgy = pcgy[icgnow] - pcgy[jcgnow];
    dcgz = pcgz[icgnow] - pcgz[jcgnow];		
    rcg = sqrt(dcgx*dcgx+dcgy*dcgy+dcgz*dcgz);

    dr = RCGBOND/rcg;

    dx = dcgx * dr + px1 - px2; 
    dy = dcgy * dr + py1 - py2; 
    dz = dcgz * dr + pz1 - pz2; 
    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r-eqbond[idx];
    
    switch(itypbond[idx]){/* fr is 1/r (dv/dr) */
    case 0:
      fr = fkbond[idx]*dr/r;
      break;
    case 1:
      mb = exp(-bmorse[idx]*dr);
      fr = (2.*fkbond[idx]*bmorse[idx]*mb*(1.-mb))/r;
      break;
    }

    dx *= fr; dy *= fr; dz *= fr;

    /* == subtract previous computed COMcounterforce  == */
    amass = 1./coords->next->amass[icgnow];
    fcgx[icgnow] -= dx*amass;  fcgy[icgnow] -= dy*amass;  fcgz[icgnow] -= dz*amass;
  }
  /* now look for SO3 bonds in healing region */
  nbonds = coords->bonds.nsel_so3hr;
  iselect = coords->bonds.isel_so3hr;
  } 
}
/*-----------------------------------------------------------------------*/
/* bonds in hybrid MD                                                    */
void fbond_hr(SIMPARMS *simparms,COORDS *coords)
{
  int i,ii,j,idx,ib,nb,nbonds,*ibond,*jbond,*idxbond,*itypbond,*iselect,*icg,lhybrid;
  long ibegin,iend;
  double dx,dy,dz,r,fr,dr,mb,fhr;
  double *px,*py,*pz,*faacg,*ds,*vds,*vcg,pot;
  double *fx,*fy,*fz;
  double *eqbond,*fkbond,*bmorse;

  px = coords->px;   py = coords->py;   pz = coords->pz;
  fx = coords->fxa;  fy = coords->fya;  fz = coords->fza;
  nbonds = coords->bonds.nselect;
  iselect = coords->bonds.iselect;
  ibond = coords->bonds.ibond;
  jbond = coords->bonds.jbond;
  idxbond = coords->bonds.idxbond;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;
  faacg = simparms->faacg + (1-coords->lcg)*simparms->ncg;
  ds = simparms->dfaacg + (1-coords->lcg)*simparms->ncg;
  vds = simparms->vds_intra + (1-coords->lcg)*simparms->ncg; 
  icg = simparms->icgunit;
  vcg = simparms->vcgintern;
  lhybrid=(simparms->icalc_type == 5);

  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);

  fr = 0.;
  for(nb=ibegin;nb<iend;nb++){  /* mind you that different processors write to the same vds[i], so doesn't work parallel */
    ib = iselect[nb];
    i = ibond[ib]; j = jbond[ib]; idx = idxbond[ib];
    dx = px[i] - px[j];  
    dy = py[i] - py[j];
    dz = pz[i] - pz[j];

    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r-eqbond[idx];
    
    switch(itypbond[idx]){/* fr is 1/r (dv/dr) */
    case 0:
      fr = fkbond[idx]*dr/r;
      pot = 0.5*fkbond[idx]*dr*dr;
      break;
    case 1:
      mb = exp(-bmorse[idx]*dr);
      pot = fkbond[idx]*(1-mb)*(1-mb);
      fr = (2.*fkbond[idx]*bmorse[idx]*mb*(1.-mb))/r;
      break;
    }
        
    /* healing region switch function added here BERND */
    if(lhybrid){
      if(coords->lcg){    /* CG-CG bonds are always scaled */
	ii = (faacg[i]>faacg[j])? i:j;
	fhr = faacg[ii];
	vds[ii] += pot*ds[ii];
      }else{              /* only AA-AA bonds over different CG atoms are scaled */
	if(icg[i]==icg[j]){
	  fhr = 1.;
	  ii = icg[i];
          vcg[ii] += pot;
	}else{
	  ii = (faacg[i]<faacg[j])? i:j;
	  fhr = faacg[ii]; 
	  vds[ii] += pot*ds[ii];
	}
      }
    }else{
      fhr = 1.;
    }	
    
    fr *= fhr;
    dx *= fr; dy *= fr; dz *= fr;
    fx[i] -= dx;  fy[i] -= dy;  fz[i] -= dz;
    fx[j] += dx;  fy[j] += dy;  fz[j] += dz;
  }
}
/*-------------------------------------------------------------------*/
void getvbond(SIMPARMS *simparms,COORDS *coords,
	      double *UI,double *WI,double WItensor[9])
{
  int i,j,idx,ib,nb,nbonds,*itypbond,*ibond,*jbond,*idxbond,*iselect,*icg,lhybrid;
  long ibegin,iend;
  double dx,dy,dz,r,dr,mb,fk,fhr,spotl,sprsl,spot=0.,sprs=0.;
  double *px,*py,*pz,*faacg;
  double *eqbond,*fkbond,*bmorse;
  double sprstensor[9],sprstl[9];

  for(i=0;i<9;i++) sprstensor[i]=0.0;
  
  px = coords->px;   py = coords->py;   pz = coords->pz;
  nbonds = coords->bonds.nselect;
  iselect = coords->bonds.iselect;
  ibond = coords->bonds.ibond;
  jbond = coords->bonds.jbond;
  idxbond = coords->bonds.idxbond;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;
  faacg = simparms->faacg + (1-coords->lcg)*simparms->ncg;
  icg = simparms->icgunit;
  lhybrid=(simparms->icalc_type == 5);
  
  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);
  
#ifdef PRINT_BOND
  printf("BONDING ENERGIES\n");
#endif
  
  for(ib=ibegin;ib<iend;ib++){
    nb = iselect[ib];
    i = ibond[nb];   j = jbond[nb];   idx = idxbond[nb];
    dx = px[i] - px[j];   
    dy = py[i] - py[j];   
    dz = pz[i] - pz[j];

    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r - eqbond[idx];
    fk = fkbond[idx];
    
    /* switch function added here BERND */
    if(lhybrid){
      if(coords->lcg){    /* CG-CG bonds are always scaled */
	fhr = MAX(faacg[i],faacg[j]);
      }else{              /* only AA-AA bonds over different CG atoms are scaled */
	if(icg[i]==icg[j]){
	  fhr = 1.;
	}else{
	  fhr = MIN(faacg[i],faacg[j]);
	}
      }
    }else{
      fhr = 1.;
    }
    fk *= fhr;	
    
    switch(itypbond[idx]){
    case 0:
      spot += .5*fk*dr*dr;
      fk = fk*dr/r;
      break;
    case 1:
      mb= 1.-exp(-bmorse[idx]*dr);
      spot += fk*mb*mb;
      fk = (2.*fk*bmorse[idx]*mb*(1.-mb))/r;
      break;
    }
    
#ifdef PRINT_BOND
    printf("bond %d (type %d) between %d-%d = %g(Kcal) =(%g A)\n",nb,idx,
	   i,j,.5*fk*dr*dr/KCAL,r);
#endif
    sprs -= fk*r*r;
    sprstensor[0] -= fk*dx*dx;
    sprstensor[1] -= fk*dy*dx;
    sprstensor[2] -= fk*dz*dx;
    
    sprstensor[3] -= fk*dx*dy;
    sprstensor[4] -= fk*dy*dy;
    sprstensor[5] -= fk*dz*dy;

    sprstensor[6] -= fk*dx*dz;
    sprstensor[7] -= fk*dy*dz;
    sprstensor[8] -= fk*dz*dz;
  }
#ifdef PARA
  MPI_Allreduce(&spot,&spotl,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  MPI_Allreduce(&sprs,&sprsl,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  MPI_Allreduce(sprstensor,sprstl,9,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
#else
  spotl = spot;
  sprsl = sprs;
  for(i=0;i<9;i++) sprstl[i] = sprstensor[i];
#endif
  *UI += spotl;
  *WI += sprsl;

  for(i=0;i<9;i++) WItensor[i] += sprstl[i];
}
/*---------------------------------------------------------------*/
double getvbond_so3(SIMPARMS *simparms,COORDS *coords,int iregion)
{
  int i,j,idx,ib,nb,nbonds,*itypbond,*ibond,*jbond,*idxbond,*iselect;
  int icg, jcg, *icgunit;
  long ibegin,iend;
  double dx,dy,dz,r,dr,mb,fk,fhr,tpot;
  double *px,*py,*pz,*pcgx,*pcgy,*pcgz,*faacg;
  double px1, py1, pz1, px2, py2, pz2;
  double *eqbond,*fkbond,*bmorse;
  
  px = coords->px;   py = coords->py;   pz = coords->pz;
  pcgx = coords->next->px;   pcgy = coords->next->py;   pcgz = coords->next->pz;
  if( iregion == 0){
    /* coarse grain region selection */
    nbonds = coords->bonds.nsel_so3cgr;
    iselect = coords->bonds.isel_so3cgr;
  }else{
    /* healing region selection */
    nbonds = coords->bonds.nsel_so3hr;
    iselect = coords->bonds.isel_so3hr;		
  }
  ibond = coords->bonds.ibond2;
  jbond = coords->bonds.jbond2;
  idxbond = coords->bonds.idxbond2;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;
  faacg = simparms->faacg + (1-coords->lcg)*simparms->ncg;
  icgunit = simparms->icgunit; 
  tpot = 0.;
  
  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);
  
#ifdef PRINT_BOND
  printf("BONDING ENERGIES\n");
#endif
  
  for(ib=ibegin;ib<iend;ib++){
    nb = iselect[ib];
    i = ibond[nb];   j = jbond[nb];   idx = idxbond[nb];
    px1 = px[i]; py1 = py[i]; pz1 = pz[i]; 
    px2 = px[j]; py2 = py[j]; pz2 = pz[j];
    icg = icgunit[i]; jcg = icgunit[j];
    
    /* renormalize for distance between CG beads */
    dx = pcgx[icg] - pcgx[jcg];  
    dy = pcgy[icg] - pcgy[jcg];
    dz = pcgz[icg] - pcgz[jcg];		
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = RCGBOND/r;

#ifdef NORESCALING
    dr = 1.0;
#endif

    dx *= dr; 
    dy *= dr; 
    dz *= dr;

    dx +=  px1 - px2; 
    dy +=  py1 - py2; 
    dz +=  pz1 - pz2;     
    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r - eqbond[idx];
    fk = fkbond[idx];
    
    fhr = 1. - MIN(faacg[i],faacg[j]);
    fk *= fhr;
    
    switch(itypbond[idx]){
    case 0:
      tpot += .5*fk*dr*dr;
      break;
    case 1:
      mb= 1.-exp(-bmorse[idx]*dr);
      tpot += fk*mb*mb;
      break;
    }
        
#ifdef PRINT_BOND
    printf("bond %d (type %d) between %d-%d = %g(Kcal) =(%g A)\n",nb,idx,
	   i,j,.5*fk*dr*dr/KCAL,r);
#endif
  }
#ifdef PARA
#else
#endif
  return tpot;
  
}
/*---------------------------------------------------------------*/
double getvbond_intern(SIMPARMS *simparms,COORDS *coords)
{
  int i,j,idx,ib,nbonds,*itypbond,*ibond,*jbond,*idxbond,*iaacg,*icg,iat;
  long ibegin,iend;
  double dx,dy,dz,r,dr,mb,fk,fr;
  double *px,*py,*pz;
  double *fx,*fy,*fz;
  double *eqbond,*fkbond,*bmorse,*spot,pot,tpot;
  
  px = coords->px;   py = coords->py;   pz = coords->pz;
  fx = coords->fxa;  fy = coords->fya;  fz = coords->fza;
  nbonds = coords->bonds.nbonds;
  ibond = coords->bonds.ibond;
  jbond = coords->bonds.jbond;
  idxbond = coords->bonds.idxbond;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;
  iaacg = simparms->iaacg;
  icg = simparms->icgunit;
  spot = simparms->vcgintern;
  tpot = 0;
  
  decomp1d((long)nbonds,simparms->size,simparms->rank,&ibegin,&iend);

#ifdef PRINT_BOND
  printf("BONDING ENERGIES\n");
#endif
  
  for(ib=ibegin;ib<iend;ib++){
    i = ibond[ib];   j = jbond[ib];
    iat = icg[i];
    if(iat!=icg[j]) continue;
    if(iaacg[iat]==0) continue;
    idx = idxbond[ib];
    dx = px[i] - px[j];   
    dy = py[i] - py[j];   
    dz = pz[i] - pz[j];
    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r - eqbond[idx];
    fk = fkbond[idx];
    
    switch(itypbond[idx]){
    case 0:
      pot = .5*fk*dr*dr;
      tpot += pot;
      spot[iat] += pot;
      fr = fk*dr/r;
      break;
    case 1:
      mb= 1.-exp(-bmorse[idx]*dr);
      pot= fk*mb*mb;
      tpot += pot;
      spot[iat] += pot;
      fr = (2.*fk*bmorse[idx]*mb*(1.-mb))/r;
      break;
    }
    dx *= fr; dy *= fr; dz *= fr;
    fx[i] -= dx;  fy[i] -= dy;  fz[i] -= dz;
    fx[j] += dx;  fy[j] += dy;  fz[j] += dz;	
  }
  return tpot;
}
/*---------------------------------------------------------------*/
/* for now I only need to recompute the forces; the potential is already stored in vcgintern */
double getvbond_intern_CG(SIMPARMS *simparms,COORDS *coords,int icg)
{
  int i,j,idx,ib,*itypbond,*ibond,*jbond,*idxbond,*ilist;
  double dx,dy,dz,r,dr,mb,fk,fr;
  double *px,*py,*pz;
  double *fx,*fy,*fz;
  double *eqbond,*fkbond,*bmorse,pot,tpot;
  
  px = coords->px;   py = coords->py;   pz = coords->pz;
  fx = coords->fxa;  fy = coords->fya;  fz = coords->fza;
  ibond = coords->bonds.ibond;
  jbond = coords->bonds.jbond;
  idxbond = coords->bonds.idxbond;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;
  ilist = coords->bonds.icgunit;
//  spot = simparms->vcgintern;
  tpot = 0;
		
#ifdef PRINT_BOND
  printf("BONDING ENERGIES\n");
#endif
  
  for(ib=ilist[icg];ib<ilist[icg+1];ib++){
    i = ibond[ib];   j = jbond[ib];
    idx = idxbond[ib];
    dx = px[i] - px[j];   
    dy = py[i] - py[j];   
    dz = pz[i] - pz[j];
    
    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r - eqbond[idx];
    fk = fkbond[idx];
    
    switch(itypbond[idx]){
    case 0:
      pot = .5*fk*dr*dr;
      tpot += pot;
      fr = fk*dr/r;
      break;
    case 1:
      mb= 1.-exp(-bmorse[idx]*dr);
      pot= fk*mb*mb;
      tpot += pot;
      fr = (2.*fk*bmorse[idx]*mb*(1.-mb))/r;
      break;
    }
    dx *= fr; dy *= fr; dz *= fr;
//    printf("BOND: %d %d %d %lf %lf %lf %lf\ns",ib,i,j,r,dx,dy,dz);
    fx[i] -= dx;  fy[i] -= dy;  fz[i] -= dz;
    fx[j] += dx;  fy[j] += dy;  fz[j] += dz;	
  }
  return tpot;
}
/*---------------------------------------------------------------*/
void fcbond(COORDS *coords,double **fcmat)
{
  int i,j,idx,nb,ii,jj,kk,nbonds;
  int *ibond,*jbond,*idxbond,*itypbond;
  double dx,dy,dz,r,r3,dr,drx,dry,drz,fk,bm,mb;
  double dxx,dyy,dzz,dxy,dxz,dyz,dudr,du2dr2;
  double *px,*py,*pz;
  double *eqbond,*fkbond,*bmorse;

  px = coords->px;   py = coords->py;   pz = coords->pz;
  nbonds = coords->bonds.nbonds;
  ibond = coords->bonds.ibond;
  jbond = coords->bonds.jbond;
  idxbond = coords->bonds.idxbond;
  itypbond = coords->bonds.itypbond;
  eqbond = coords->bonds.eqbond;
  fkbond = coords->bonds.fkbond;
  bmorse = coords->bonds.bmorse;

  dudr = 0.;
  du2dr2 = 0.;
  for(nb=0;nb<nbonds;nb++){
    i = ibond[nb];  j = jbond[nb];  idx = idxbond[nb];

    dx = px[i] - px[j];
    dy = py[i] - py[j];
    dz = pz[i] - pz[j];

    r = sqrt(dx*dx+dy*dy+dz*dz);
    dr = r-eqbond[idx];
    fk = fkbond[idx];
    
    switch(itypbond[idx]){
    case 0:
      dudr = fk*dr;
      du2dr2 = fk;
      break;
    case 1:
      bm = bmorse[idx];
      mb = exp(-bm*dr);
      dudr = 2.*fk*bm*mb*(1.-mb);
      du2dr2 = (-2.*fk*bm*bm*mb)*(1.-2.*mb);
      break;
    }
    
    r = 1./r;     r3 = (r*r*r);
    drx=dx*r;    dry=dy*r;    drz=dz*r;
    dxx = dx*dx*r3-r;
    dyy = dy*dy*r3-r;
    dzz = dz*dz*r3-r;
    dxy = dx*dy*r3;
    dxz = dx*dz*r3;
    dyz = dy*dz*r3;

    dxx = du2dr2*drx*drx - dudr*dxx;
    dxy = du2dr2*drx*dry - dudr*dxy;
    dxz = du2dr2*drx*drz - dudr*dxz;
    dyy = du2dr2*dry*dry - dudr*dyy;
    dyz = du2dr2*dry*drz - dudr*dyz;
    dzz = du2dr2*drz*drz - dudr*dzz;

    ii=i*3-1;
    jj=j*3-1;

    /* fill in diagonal elements of matrix  */

    fcmat[ii+1][ii+1] += dxx;
    fcmat[ii+1][ii+2] += dxy;
    fcmat[ii+1][ii+3] += dxz;
    fcmat[ii+2][ii+1] += dxy;
    fcmat[ii+2][ii+2] += dyy;
    fcmat[ii+2][ii+3] += dyz;
    fcmat[ii+3][ii+1] += dxz;
    fcmat[ii+3][ii+2] += dyz;
    fcmat[ii+3][ii+3] += dzz;
    
    fcmat[jj+1][jj+1] += dxx;
    fcmat[jj+1][jj+2] += dxy;
    fcmat[jj+1][jj+3] += dxz;
    fcmat[jj+2][jj+1] += dxy;
    fcmat[jj+2][jj+2] += dyy;
    fcmat[jj+2][jj+3] += dyz;
    fcmat[jj+3][jj+1] += dxz;
    fcmat[jj+3][jj+2] += dyz;
    fcmat[jj+3][jj+3] += dzz;

      /*  Fill in off-diagonal elements  */

    if(ii>jj) {kk=ii;ii=jj;jj=kk;}
    fcmat[ii+1][jj+1] -= dxx;
    fcmat[ii+1][jj+2] -= dxy;
    fcmat[ii+1][jj+3] -= dxz;
    fcmat[ii+2][jj+1] -= dxy;
    fcmat[ii+2][jj+2] -= dyy;
    fcmat[ii+2][jj+3] -= dyz;
    fcmat[ii+3][jj+1] -= dxz;
    fcmat[ii+3][jj+2] -= dyz;
    fcmat[ii+3][jj+3] -= dzz;
  }
}

/*-----------------------------------------------------------------------*/
void fcbondn(COORDS *coords,double **fcmat)
{
  int nb,idx;

  for(nb=0;nb<coords->bonds.nbonds;nb++){

    ibon = coords->bonds.ibond[nb];
    jbon = coords->bonds.jbond[nb];
    idx = coords->bonds.idxbond[nb];

    itbon = coords->bonds.itypbond[idx];
    eqbon = coords->bonds.eqbond[idx];
    fkbon = coords->bonds.fkbond[idx];
    bmbon = coords->bonds.bmorse[idx];

    ngradv2(coords->px,coords->py,coords->pz,ibon,ibon,potbond,fcmat);
    ngradv2(coords->px,coords->py,coords->pz,ibon,jbon,potbond,fcmat);
    ngradv2(coords->px,coords->py,coords->pz,jbon,ibon,potbond,fcmat);
    ngradv2(coords->px,coords->py,coords->pz,jbon,jbon,potbond,fcmat);
  }
}
/*----------------------------------------------------------------*/
double potbond(double *px,double *py,double *pz)
{
  double dx,dy,dz,r,dr,spot;

  dx = px[ibon] - px[jbon];
  dy = py[ibon] - py[jbon];
  dz = pz[ibon] - pz[jbon];
  r = sqrt(dx*dx+dy*dy+dz*dz);

  dr = r-eqbon;
  spot = 0.;
  switch(itbon){
  case 0:
    spot = .5*fkbon*dr*dr;
    break;
  case 1:
    r = 1.-exp(-bmbon*dr);
    spot = fkbon*r*r;
    break;
  }
  
  return spot;
}
/*------------------------------------------------------------------*/
