/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* read in the parameter files */

#include "md.h"

/* #define PRINT_ATOMS */


/*--------------------------------------------------------------  */
/* Check if atoms involved in bond, bend, tors, etc, interactions */
/* for the SO3 dynamics belong to different CG particles (would   */
/* be a waste to compute them otherwise and probably an error).   */
void check_so3params(SIMPARMS *simparms,COORDS *coords)
{
  int i,nb,iat,jat,kat,icg,*ib,*jb,*kb,*lb;
  int *icgunit = simparms->icgunit;
  int ncg = simparms->ncg;
  LINE line;
  
  /* check if we are doing hybrid MD */
  if( simparms->icalc_type != 5){
    if( coords->bonds.nbonds2 != 0){ 
      md_warning("Found non-zero number of SO3 rotation bonds in topfile while not in hybrid MD. Setting bonds2 to zero");
      coords->bonds.nbonds2 = 0;
    }
    if( coords->bends.nbends2 != 0){ 
      md_warning("Found non-zero number of SO3 rotation bends in topfile while not in hybrid MD. Setting bends2 to zero");
      coords->bends.nbends2 = 0;
    }
    if( coords->tors.ntorss2 != 0){ 
      md_warning("Found non-zero number of SO3 rotation torsions in topfile while not in hybrid MD. Setting ntorss2 to zero");
      coords->tors.ntorss2 = 0;
    }
    if( coords->onfo.n14s2 != 0){ 
      md_warning("Found non-zero number of SO3 rotation one-fours in topfile while not in hybrid MD. Setting n14s2 to zero");
      coords->onfo.n14s2 = 0;
    }
    if( coords->bondxs.nbondxs2 != 0){ 
      md_warning("Found non-zero number of SO3 rotation cross-bonds in topfile while not in hybrid MD. Setting nbondxs2 to zero");
      coords->bondxs.nbondxs2 = 0;
    }
    return;
  }
  
  /* check SO3 bonds */
  nb=coords->bonds.nbonds2;
  ib=coords->bonds.ibond2;
  jb=coords->bonds.jbond2;
  for(i=0;i<nb;i++){
    iat = ib[i]-ncg;
    icg = icgunit[iat];
    iat = jb[i]-ncg;
    if( icg == icgunit[iat] ){
      sprintf(line,"both atoms involved in SO3 bond no.%d belong to the same CG blob no.%d ",i,icg);
      md_error(line);
    }
  }
  
  /* check SO3 bends */
  nb=coords->bends.nbends2;
  ib=coords->bends.ibend2;
  jb=coords->bends.jbend2;
  kb=coords->bends.kbend2;
  for(i=0;i<nb;i++){
    iat = ib[i]-ncg;
    icg = icgunit[iat];
    iat = jb[i]-ncg;
    jat = kb[i]-ncg;
    if( (icg == icgunit[iat]) && (icg == icgunit[jat]) ){
      sprintf(line,"all atoms involved in SO3 bend no.%d belong to the same CG blob no.%d ",i,icg);
      md_error(line);
    }
  }
  
  /* check SO3 tors */
  nb=coords->tors.ntorss2;
  ib=coords->tors.itors2;
  jb=coords->tors.jtors2;
  kb=coords->tors.ktors2;
  lb=coords->tors.ltors2;
  for(i=0;i<nb;i++){
    iat = ib[i]-ncg;
    icg = icgunit[iat];
    iat = jb[i]-ncg;
    jat = kb[i]-ncg;
    kat = lb[i]-ncg;
    if( (icg == icgunit[iat]) && (icg == icgunit[jat]) && (icg == icgunit[kat])){
      sprintf(line,"all atoms involved in SO3 torsion no.%d belong to the same CG blob no.%d ",i,icg);
      md_error(line);
    }
  }
  
  /* check SO3 onefour */
  nb=coords->onfo.n14s2;
  ib=coords->onfo.i142;
  jb=coords->onfo.j142;
  for(i=0;i<nb;i++){
    iat = ib[i]-ncg;
    icg = icgunit[iat];
    iat = jb[i]-ncg;
    if( icg == icgunit[iat] ){
      sprintf(line,"both atoms involved in SO3 onefour no.%d belong to the same CG blob no.%d ",i,icg);
      md_error(line);
    }
  }
  
  /* check SO3 bondxs */
  nb=coords->bondxs.nbondxs2;
  ib=coords->bondxs.ibondx2;
  jb=coords->bondxs.jbondx2;
  kb=coords->bondxs.kbondx2;
  for(i=0;i<nb;i++){
    iat = ib[i]-ncg;
    icg = icgunit[iat];
    iat = jb[i]-ncg;
    jat = kb[i]-ncg;
    if( (icg == icgunit[iat]) && (icg == icgunit[jat]) ){
      sprintf(line,"all atoms involved in SO3 crossbond no.%d belong to the same CG blob no.%d ",i,icg);
      md_error(line);
    }
  }
  
}


/*--------------------------------------------------------------*/

void read_parmfile(SIMPARMS *simparms,COORDS *coords,STARTUP *startup,
		   INTER *inter)
{
  int i,j,k,ispec,jspec,ntypes,ncgtypes,ioff,ii,icg,ifound,lhybrid;
  int iatom,ngroup,natoms,ngpm,ncg,naa,ncgspec,*iorder;
  int **ibo,**jbo,**ibl; WORD **tbo;
  int **ibe,**jbe,**kbe,**tybe,**ibel; WORD **tbe;
/*  int **ibeq,**jbeq,**kbeq; WORD **tbeq; */
  int **ito,**jto,**kto,**lto,**itl; WORD **tto;
  int **iof,**jof,**iol; WORD **tof;
  int **ibonx,**jbonx,**kbonx,**ibonxl; WORD **tbonx;
  int *atm_idx;
  WORD tgroup;
  SPECIES *spec_now, *spec_tmp, *spec_next;
  ATOM_TOPOL **atom_topol;
  LINE line;
  
  lhybrid = ((simparms->icalc_type == 5) || (simparms->icalc_type == 6) );
  
  /* allocate temparary memory */
  atom_topol = (ATOM_TOPOL **)cmalloc(startup->nspec*sizeof(ATOM_TOPOL *));

  /* bonds */
  ibo = (int **)cmalloc(startup->nspec*sizeof(int*));
  jbo = (int **)cmalloc(startup->nspec*sizeof(int*));
  tbo = (WORD **)cmalloc(startup->nspec*sizeof(WORD*));
  ibl = (int **)cmalloc(startup->nspec*sizeof(int*));

  /* bends */
  ibe = (int **)cmalloc(startup->nspec*sizeof(int*));
  jbe = (int **)cmalloc(startup->nspec*sizeof(int*));
  kbe = (int **)cmalloc(startup->nspec*sizeof(int*));
  tybe = (int **)cmalloc(startup->nspec*sizeof(int*));
  tbe = (WORD **)cmalloc(startup->nspec*sizeof(WORD*));
  ibel = (int **)cmalloc(startup->nspec*sizeof(int*));

  /* quartic bends */
/*
  ibeq = (int **)cmalloc(startup->nspec*sizeof(int*));
  jbeq = (int **)cmalloc(startup->nspec*sizeof(int*));
  kbeq = (int **)cmalloc(startup->nspec*sizeof(int*));
  tbeq = (WORD **)cmalloc(startup->nspec*sizeof(WORD*));
*/
  /* torsions */
  ito = (int **)cmalloc(startup->nspec*sizeof(int*));
  jto = (int **)cmalloc(startup->nspec*sizeof(int*));
  kto = (int **)cmalloc(startup->nspec*sizeof(int*));
  lto = (int **)cmalloc(startup->nspec*sizeof(int*));
  tto = (WORD **)cmalloc(startup->nspec*sizeof(WORD*));
  itl = (int **)cmalloc(startup->nspec*sizeof(int*));

  /* onefour */
  iof = (int **)cmalloc(startup->nspec*sizeof(int*));
  jof = (int **)cmalloc(startup->nspec*sizeof(int*));
  tof = (WORD **)cmalloc(startup->nspec*sizeof(WORD*));
  iol = (int **)cmalloc(startup->nspec*sizeof(int*));

  /* cross bonds */
  ibonx = (int **)cmalloc(startup->nspec*sizeof(int*));
  jbonx = (int **)cmalloc(startup->nspec*sizeof(int*));
  kbonx = (int **)cmalloc(startup->nspec*sizeof(int*));
  tbonx = (WORD **)cmalloc(startup->nspec*sizeof(WORD*));
  ibonxl = (int **)cmalloc(startup->nspec*sizeof(int*));

  /* read in all connectivity data from files */
  simparms->naa = simparms->ncg = 0;
  read_top_file(simparms,startup,atom_topol,
		ibo,jbo,ibl,tbo,ibe,jbe,kbe,tybe,ibel,tbe,
		ito,jto,kto,lto,itl,tto,iof,jof,iol,tof,
		ibonx,jbonx,kbonx,ibonxl,tbonx);
  
#ifdef DEBUG
  printf("Total number of atoms = %d\n",simparms->natoms);
#endif
  natoms = coords->natoms = simparms->natoms;
  iorder = (int *)cmalloc(startup->nspec*sizeof(int));   /* holds order of species after CG-AA reordering */
  if( lhybrid ){
    ncg = simparms->ncg;
    if(ncg==0) ncg=1;
    naa = simparms->naa;
    /* reorder species in startup struct */
    spec_tmp = (SPECIES *)cmalloc(startup->nspec*sizeof(SPECIES));
    spec_tmp[0]=startup->spec_root;
    for(ispec=1;ispec<startup->nspec;ispec++)
       spec_tmp[ispec]=*spec_tmp[ispec-1].next;
    ifound=0;
    spec_now = &startup->spec_root;
    for(ispec=0;ispec<startup->nspec;ispec++){
        /* first get the CG species */
      if(atom_topol[ispec][0].icgflag == 1){
        *spec_now = spec_tmp[ispec];                /* overwrite */
        spec_now->next = spec_tmp[ifound].next;     /* now restore the link to next */
        spec_now = spec_now->next;                  /* then go to next and repeat */
        iorder[ifound]=ispec;
        ifound++;
      }
    }
    ncgspec=ifound;
    ii=0;
       /* then get the AA species in the same molecule name order as the CG species */
    for(i=0,spec_next = &startup->spec_root;i<ncgspec;i++,spec_next=spec_next->next){
      for(ispec=0;ispec<startup->nspec;ispec++){
        if( (strcmp(spec_next->name,spec_tmp[ispec].name) == 0) &&
            (atom_topol[ispec][0].icgflag == 0) ){
          *spec_now = spec_tmp[ispec];                /* overwrite */
          spec_now->next = spec_tmp[ifound].next;     /* now restore the link to next */
          iorder[ifound]=ispec;
          ifound++;
          ii++;
        }
      }
      if(ii!=1) {
        md_error("Failed to pair all CG species with all AA species.\nCheck the matching of molecule names; should be unique per AA-CG pair...");
      }
      spec_now = spec_now->next;                  /* then go to next and repeat */
      ii=0;
    }
    if(ifound != startup->nspec){
      md_error("Failed to pair all CG species with all AA species.\nCheck the matching of molecule names; should be unique per AA-CG pair...");
    }
  }else{ /* simparms->icalc_type != 5 and != 6 */
    ncg = simparms->ncg=0;
	simparms->naa = simparms->natoms;
    for(ispec=0;ispec<startup->nspec;ispec++) iorder[ispec]=ispec;  /* no reordering otherwise */ 
  }
  allocate_coords(simparms,coords);

  /* allocate memory for maps and charges and masses */
  simparms->itype = (int *)cmalloc(5*natoms*sizeof(int));
  simparms->iatom     = simparms->itype +   natoms;
  simparms->imolecule = simparms->itype + 2*natoms;
  simparms->ispecies  = simparms->itype + 3*natoms;
  simparms->igroup    = simparms->itype + 4*natoms;
    
  simparms->atom = (WORD *)cmalloc(natoms*sizeof(WORD));
  simparms->group = (WORD *)cmalloc(natoms*sizeof(WORD));
  simparms->mole = (WORD *)cmalloc(startup->nspec*sizeof(WORD));
  simparms->nspec = startup->nspec;

  simparms->mem_bytes += natoms*(6*sizeof(int)+2*sizeof(WORD));
  simparms->mem_bytes += startup->nspec*sizeof(WORD);

  if(lhybrid){
    simparms->icgunit = (int *)cmalloc(naa*sizeof(int));
    simparms->napcg = (int *)cmalloc(ncg*sizeof(int));
    for(icg=0;icg<ncg;++icg) simparms->napcg[icg]=0;
    simparms->icglist = (int **)cmalloc(ncg*sizeof(int *));
    simparms->icglist[0] = (int *)cmalloc(naa*sizeof(int));
  }
  
  atm_idx = (int *)cmalloc(natoms*sizeof(int));

  /* get all the atom types */
  ngroup = ntypes = ncgtypes = iatom = 0;

  /* fill in the maps */
  for(ispec=0,spec_now=&startup->spec_root;ispec < startup->nspec;
      ispec++,spec_now=spec_now->next){
    jspec = iorder[ispec];
    /* save molecule type */
    strcpy((simparms->mole)[ispec],spec_now->name);

    /* set up to check for group type */
    strcpy((simparms->group)[ngroup],atom_topol[jspec][0].group);
    strcpy(tgroup,atom_topol[jspec][0].group);
    ngpm = 0;
      
    for(j=0;j<spec_now->napm;j++){
	
      /* check for new atom type */
      if( (ncgtypes == 0) && ((atom_topol[jspec][0].icgflag) == 0) && (lhybrid)) 
	ncgtypes=ntypes;  /* if CG type equals a AA type -> add type as new */
      for(i=ncgtypes;i<ntypes;i++){
        if(strcasecmp(atom_topol[jspec][j].type,(simparms->atom)[i]) == 0) {
          break;
        }
      }
      if(i==ntypes){
        strcpy((simparms->atom)[ntypes],atom_topol[jspec][j].type);
        ntypes++;
      }
	
      /* check for new group type */
      if(strcasecmp(atom_topol[jspec][j].group,tgroup) != 0) {
        strcpy(tgroup,atom_topol[jspec][j].group);
        ngpm++;
      }
	
      for(k=0;k<spec_now->nmol;k++){
        ioff = iatom + j + k*spec_now->napm;
        //	  atom_topol[ispec][j].atm_idx
        simparms->itype[ioff]     = i;
        simparms->iatom[ioff]     = j;
        simparms->imolecule[ioff] = k;
        simparms->ispecies[ioff]  = ispec;
        simparms->igroup[ioff]    = ngpm+ngroup+k;
        atm_idx[ioff]             = atom_topol[jspec][j].atm_idx;
        (coords->amass)[ioff] = atom_topol[jspec][j].mass*MCONV;
        (coords->qch)[ioff]   = (atom_topol[jspec][j].charge*sqrt(CCONV)*
				 simparms->scalecharge);
        (coords->alpha)[ioff] = atom_topol[jspec][j].alpha;
        strcpy((simparms->group)[ngpm+ngroup+k],atom_topol[jspec][j].group);
      }
    }
    iatom += spec_now->napm*spec_now->nmol;
    ngroup += ngpm*spec_now->nmol;
  }
  /* reallocate atom_types to just the number needed */
  simparms->atom = (WORD *)realloc(simparms->atom,ntypes*sizeof(WORD));
  
  
  /* cross reference all atom and coarse grain sites */
  if(lhybrid){
    for(iatom=ncg;iatom<natoms;++iatom){
      ifound=0;
      for(icg=0;icg<ncg;++icg){
	if(strcasecmp((simparms->mole)[simparms->ispecies[iatom]],(simparms->mole)[simparms->ispecies[icg]]) == 0){
	  if((simparms->imolecule)[iatom] == (simparms->imolecule)[icg]){
            jspec=iorder[simparms->ispecies[iatom]];
	    if(strcasecmp(atom_topol[jspec][simparms->iatom[iatom]].cg_typ,(simparms->atom)[simparms->itype[icg]]) == 0){
	      if(atom_topol[jspec][simparms->iatom[iatom]].cg_ind == atm_idx[icg]){
		simparms->icgunit[iatom-ncg]=icg;
		ifound=1;
		simparms->napcg[icg]++;
		break;
	      }
	    }
	  }
	}
      }
      if(ifound == 0){
	if(simparms->icalc_type == 5){
	  fprintf(stderr,"ERR: Did not find CG unit for molecule=%s, cgtyp=%s, cgind=%d\n",(simparms->mole)[simparms->ispecies[iatom]],atom_topol[simparms->ispecies[iatom]][simparms->iatom[iatom]].cg_typ,atom_topol[simparms->ispecies[iatom]][simparms->iatom[iatom]].cg_ind);
          exit(1);
	}
	simparms->icgunit[iatom-ncg]=-1;
      }
    }
    ioff=0;
    for(icg=0;icg<ncg;++icg){
      simparms->icglist[icg] = simparms->icglist[0] + ioff;
      for(iatom=0;iatom<naa;++iatom){
    	if(simparms->icgunit[iatom]==icg){
	  simparms->icglist[0][ioff]=iatom;
	  ioff++;
	}
      }
    }
  }
  
  /* check for CG units without AA atoms */
  for(icg=0;icg<ncg;icg++){
    if(simparms->napcg[icg] == 0){
      jspec=iorder[simparms->ispecies[icg]];
      iatom = simparms->iatom[icg];
      sprintf(line,"CG site number %d (index %d and type %s in molecule %s) has zero atoms!",icg,
	      atom_topol[jspec][iatom].atm_idx,atom_topol[jspec][iatom].type,(simparms->mole)[simparms->ispecies[icg]]);
      md_error(line);
    }
  }

  /* check if any masses are set to 0.1 while not in hybrid MD */
  if(!lhybrid){
    for(i=0;i<natoms;i++){
      if(coords->amass[i] == 0.1*MCONV){
	sprintf(line,"Atom %d has mass equil to 0.1 while not in hybrid MD",i);
	md_error(line);
      }
    }
  }
  
  /* select all atoms by default */
  coords->nselect =  simparms->natoms;
  coords->iselect = (int *)cmalloc(natoms*sizeof(int));
  for(i=0;i<natoms;i++) coords->iselect[i]=i;
  
#ifdef PRINT_ATOMS
  printf(" i  ispec imol igrp iatom ityp  idx  icg spec typ grp mass q \n");
  for(i=0;i<natoms;i++){
    k=i-ncg;
    if(i<ncg) {j=-1;} else {j=simparms->icgunit[k];}
    printf("%3d %4d %4d %4d %4d %4d %4d %4d %s %s %s %lg %lg\n",
	   i,simparms->ispecies[i],simparms->imolecule[i],simparms->igroup[i],
	   simparms->iatom[i],simparms->itype[i],
	   atm_idx[i],j,
	   (simparms->mole)[simparms->ispecies[i]],
	   (simparms->atom)[simparms->itype[i]],
	   (simparms->group)[simparms->igroup[i]],
	   (coords->amass)[i],(coords->qch)[i]);
  }
#endif

  simparms->ntypes = inter->ntype = ntypes;
  simparms->ncgtypes = inter->ncgtype = ncgtypes;
  if(simparms->rank==0){
    sprintf(line,"There are %d groups and %d types of atoms",ngroup,ntypes);
    md_stdout(line);
  }
  /* let the intermolecular force routine know the atom types */
  inter->itype = (int *)cmalloc(natoms*sizeof(int));
  for(i=0;i<natoms;i++){inter->itype[i] = simparms->itype[i];}

  /* search the data base for the interactions */
  search_ter_base(simparms,startup,inter);
  
  /* get the types of intra interactions */

  search_bond_base(simparms,startup->bond_file,ibo,jbo,ibl,tbo,
		   atom_topol,startup->spec_root,startup->nspec,
		   &coords->bonds,iorder);
  search_bend_base(simparms,startup->bend_file,ibe,jbe,kbe,tybe,ibel,tbe,atom_topol,
		   startup->spec_root,startup->nspec,&coords->bends,iorder);
/*
  search_bendquart_base(simparms,startup->bend_file,ibeq,jbeq,kbeq,tbeq,atom_topol,
		   startup->spec_root,startup->nspec,&coords->bendquarts);
*/
  search_tors_base(simparms,startup->tors_file,ito,jto,kto,lto,itl,tto,atom_topol,
		   startup->spec_root,startup->nspec,&coords->tors,iorder);
  search_onfo_base(simparms,startup->onfo_file,simparms->icalc_type,
		   iof,jof,iol,tof,atom_topol,
		   startup->spec_root,startup->nspec,startup->scale_onfo,
		   startup->scale_onfo_e,inter->ntable,&coords->onfo,iorder);

  /* cross potential interactions */
  search_bocross_base(simparms,startup->bond_file,ibonx,jbonx,kbonx,ibonxl,tbonx,
		      atom_topol,startup->spec_root,startup->nspec,
		      &coords->bondxs,iorder);

  /* setup the charge charge interaction  */

  set_charges(simparms,coords,inter,startup);

  /* setup exclusions */
  set_exclude(simparms,coords,inter,atm_idx);
  
  check_so3params(simparms,coords);
  
  /* free temporary storage */
  for(i=0;i<startup->nspec;i++){
    free(atom_topol[i]);
    if(ibo[i]!=NULL){free(ibo[i]); free(jbo[i]); free(ibl[i]); free(tbo[i]);}
    if(ibe[i]!=NULL){
      free(ibe[i]); free(jbe[i]); free(kbe[i]); free(tybe[i]); free(ibel[i]); free(tbe[i]);
    }
    /* if(ibeq[i]!=NULL){free(ibeq[i]); free(jbeq[i]); free(kbeq[i]); free(tbeq[i]);} */
     if(ito[i]!=NULL){
      free(ito[i]); free(jto[i]); free(kto[i]); free(lto[i]); free(itl[i]); free(tto[i]);
    }
    if(iof[i]!=NULL){free(iof[i]); free(jof[i]); free(iol[i]); free(tof[i]);}
    if(ibonx[i]!=NULL){
      free(ibonx[i]);free(jbonx[i]);free(kbonx[i]);free(ibonxl[i]);free(tbonx[i]); 
    }
  }
  free(atom_topol);
  free(ibo); free(jbo); free(ibl); free(tbo);
  free(ibe); free(jbe); free(kbe); free(tbe); free(ibel); free(tybe);
  /*  free(ibeq); free(jbeq); free(kbeq); free(tbeq); */
  free(ito); free(jto); free(kto); free(lto); free(itl); free(tto);
  if(iof != NULL) free(iof); if(jof != NULL) free(jof); if(iol != NULL) free(iol); if(tof != NULL) free(tof);
  free(ibonx);free(jbonx);free(kbonx);free(ibonxl);free(tbonx);
  if(lhybrid)free(spec_tmp);
  free(iorder);
}

/*------------------------------------------------------------------------*/
void read_top_file(SIMPARMS *simparms,STARTUP *startup,
		   ATOM_TOPOL **atom_topol,
		   int **ibo,int **jbo,int **ibl, WORD **tbo, 
		   int **ibe,int **jbe,int **kbe, int **tybe,int **ibel,WORD **tbe,
		   int **ito,int **jto,int **kto,int **lto,int **itl,WORD **tto,
		   int **iof,int **jof,int **iol, WORD **tof,
		   int **ibonx,int **jbonx,int **kbonx,int **ibonxl,WORD **tbonx)
{
  int i,j,ispec;
  
  FILE *fp;
  SPECIES *spec_now;
  
  simparms->natoms = 0;
  spec_now = &(startup->spec_root);

  for(ispec=0;ispec<startup->nspec;ispec++,spec_now=spec_now->next){
    
    /* open parameter file */
    fp = fopen(spec_now->filename,"r");
    if (fp == NULL){
      fprintf(stderr,"ERROR: can't open set file (%s)\n",spec_now->filename);
      exit(1);
    } else {
      if(simparms->rank==0){
	fprintf(stdout,"Reading in parameter from file %s\n",
		spec_now->filename);
      }
    }

    get_species(fp,spec_now);

    if(spec_now->napm <= 0)
      fprintf(stderr,"ERROR, %s has %d atoms!\n",
	      spec_now->filename,spec_now->napm);

    atom_topol[ispec] = (ATOM_TOPOL *)cmalloc(spec_now->napm*
					     sizeof(atom_topol[0][0]));
    get_atom(fp,spec_now->filename,spec_now->napm,atom_topol[ispec]);

    /* bonding */
    ibo[ispec] = jbo[ispec] = ibl[ispec] = NULL;
    tbo[ispec] = NULL;
    if(spec_now->nbond>0) {
      ibo[ispec] = (int *)cmalloc(spec_now->nbond*sizeof(int));
      jbo[ispec] = (int *)cmalloc(spec_now->nbond*sizeof(int));
      ibl[ispec] = (int *)cmalloc(spec_now->nbond*sizeof(int));
      tbo[ispec] = (WORD *)cmalloc(spec_now->nbond*sizeof(WORD));
      
      get_bond(fp,spec_now->filename,spec_now->nbond,
	       ibo[ispec],jbo[ispec],ibl[ispec],tbo[ispec]);
      for(i=0;i<spec_now->nbond;i++){
				for(j=0;j<spec_now->napm;j++){
					if(atom_topol[ispec][j].atm_idx == ibo[ispec][i]) ibo[ispec][i] = j;
					if(atom_topol[ispec][j].atm_idx == jbo[ispec][i]) jbo[ispec][i] = j;
				}
      }
    }

    /* bending */
    ibe[ispec] = jbe[ispec] = kbe[ispec] = ibel[ispec] = NULL;
    tybe[ispec] = NULL; tbe[ispec] = NULL; 
    if(spec_now->nbend>0){
      ibe[ispec] = (int *)cmalloc(spec_now->nbend*sizeof(int));
      jbe[ispec] = (int *)cmalloc(spec_now->nbend*sizeof(int));
      kbe[ispec] = (int *)cmalloc(spec_now->nbend*sizeof(int));
      tybe[ispec] = (int *)cmalloc(spec_now->nbend*sizeof(int));
      ibel[ispec] = (int *)cmalloc(spec_now->nbend*sizeof(int));
      tbe[ispec] = (WORD *)cmalloc(spec_now->nbend*sizeof(WORD));
      get_bend(fp,spec_now->filename,spec_now->nbend,
	       ibe[ispec],jbe[ispec],kbe[ispec],tybe[ispec],ibel[ispec],tbe[ispec]);
      for(i=0;i<spec_now->nbend;i++){
				for(j=0;j<spec_now->napm;j++){
					if(atom_topol[ispec][j].atm_idx == ibe[ispec][i]) ibe[ispec][i] = j;
					if(atom_topol[ispec][j].atm_idx == jbe[ispec][i]) jbe[ispec][i] = j;
					if(atom_topol[ispec][j].atm_idx == kbe[ispec][i]) kbe[ispec][i] = j;
				}
      }
    }

    /* torsions */
    ito[ispec] = jto[ispec] = kto[ispec] = lto[ispec] = itl[ispec] =NULL;
    tto[ispec] = NULL;
    if(spec_now->ntors>0) {
      ito[ispec] = (int *)cmalloc(spec_now->ntors*sizeof(int));
      jto[ispec] = (int *)cmalloc(spec_now->ntors*sizeof(int));
      kto[ispec] = (int *)cmalloc(spec_now->ntors*sizeof(int));
      lto[ispec] = (int *)cmalloc(spec_now->ntors*sizeof(int));
      itl[ispec] = (int *)cmalloc(spec_now->ntors*sizeof(int));
      tto[ispec] = (WORD *)cmalloc(spec_now->ntors*sizeof(WORD));
      get_tors(fp,spec_now->filename,spec_now->ntors,
	       ito[ispec],jto[ispec],kto[ispec],lto[ispec],itl[ispec],tto[ispec]);
      for(i=0;i<spec_now->ntors;i++){
				for(j=0;j<spec_now->napm;j++){
					if(atom_topol[ispec][j].atm_idx == ito[ispec][i]) ito[ispec][i] = j;
					if(atom_topol[ispec][j].atm_idx == jto[ispec][i]) jto[ispec][i] = j;
					if(atom_topol[ispec][j].atm_idx == kto[ispec][i]) kto[ispec][i] = j;
					if(atom_topol[ispec][j].atm_idx == lto[ispec][i]) lto[ispec][i] = j;
				}
      }
    }
    /* onefour */
    iof[ispec] = jof[ispec] = iol[ispec] = NULL; tof[ispec] = NULL;
    if(spec_now->n14>0){
      iof[ispec] = (int *)cmalloc(spec_now->n14*sizeof(int));
      jof[ispec] = (int *)cmalloc(spec_now->n14*sizeof(int));
      iol[ispec] = (int *)cmalloc(spec_now->n14*sizeof(int));
      tof[ispec] = (WORD *)cmalloc(spec_now->n14*sizeof(WORD));
      get_onefour(fp,spec_now->filename,spec_now->n14,iof[ispec],jof[ispec],iol[ispec],tof[ispec]);
      for(i=0;i<spec_now->n14;i++){
				for(j=0;j<spec_now->napm;j++){
					if(atom_topol[ispec][j].atm_idx == iof[ispec][i]) iof[ispec][i] = j;
					if(atom_topol[ispec][j].atm_idx == jof[ispec][i]) jof[ispec][i] = j;
				}
      }
    }
    /* cross bonding */
    ibonx[ispec] = jbonx[ispec] = kbonx[ispec] = ibonxl[ispec] = NULL;
    tbonx[ispec] = NULL;
    if(spec_now->nbondx>0){
      ibonx[ispec] = (int *)cmalloc(spec_now->nbondx*sizeof(int));
      jbonx[ispec] = (int *)cmalloc(spec_now->nbondx*sizeof(int));
      kbonx[ispec] = (int *)cmalloc(spec_now->nbondx*sizeof(int));
      ibonxl[ispec] = (int *)cmalloc(spec_now->nbondx*sizeof(int));
      tbonx[ispec] = (WORD *)cmalloc(spec_now->nbondx*sizeof(WORD));
      get_bondx(fp,spec_now->filename,spec_now->nbondx,
		ibonx[ispec],jbonx[ispec],kbonx[ispec],ibonxl[ispec],tbonx[ispec]);
      for(i=0;i<spec_now->nbondx;i++){
				for(j=0;j<spec_now->napm;j++){
					if(atom_topol[ispec][j].atm_idx==ibonx[ispec][i]) ibonx[ispec][i]=j;
					if(atom_topol[ispec][j].atm_idx==jbonx[ispec][i]) jbonx[ispec][i]=j;
					if(atom_topol[ispec][j].atm_idx==kbonx[ispec][i]) kbonx[ispec][i]=j;
				}
      }
    }
    fclose(fp);
    
    simparms->natoms += spec_now->napm*spec_now->nmol;
    if(atom_topol[ispec][0].icgflag==0){  /* if true: all atom topology */
      simparms->naa += spec_now->napm*spec_now->nmol;
    }else{                                /*      coarse grain topology */
      simparms->ncg += spec_now->napm*spec_now->nmol;
    }
  }
}
