/* 
   Copyright (C) 1997 1998 1999 2000 Dr. Preston B. Moore

Dr. Preston B. Moore
Associate Director, Center for Molecular Modeling (CMM)
University of Pennsylvania, Department of Chemistry, Box 188 
231 S. 34th St. Philadelphia, PA 19104-6323 USA
EMAIL: moore@cmm.chem.upenn.edu  
WWW: http://www.cmm.upenn.edu/~moore

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* routine to read in the positions for initfile */

#include "md.h"
/* #define DEBUG */

/*------------------------------------------------------------------*/
/* read atom positions for a molecule from a library (.lib) file in */
/* a hybrid MD simulation                                           */
void readaapos(char *molename,double *px,double *py,double *pz,int natoms){

  int i,n;
  WORD filename;
  LINE line;
  FILE *fp;

  sprintf(filename,"%s.lib",molename);
  fp = fopen(filename,"r");
  if(fp == NULL){
    sprintf(line,"failed to open %s\n",filename);
    md_error(line);
  }
  
  fscanf(fp,"%d",&n);
  if(n != natoms){
    sprintf(line,"Number of atoms in %s (%d) not equal to number expected (%d)\n",filename,n,natoms);
	md_error(line);
  }
  
  for(i=0;i<natoms;i++) fscanf(fp,"%lf%lf%lf",&px[i],&py[i],&pz[i]);
}

/*------------------------------------------------------------------  */
/* read positions from a configuration file in case of a cold start   */
/* or, in case of a restart, read from the restart file the positions,*/
/* velocities, thermostat and barostat states and other accumulators. */
void readpos(char *command,char *initfile,SIMPARMS *simparms,
	     int *estep,ENERGIES *energies,COORDS *coords,SO3 *so3)
{
  char line[MAXLINELEN];
  int i,j,n,nt,nc,nb,ist,iat,icg,imol,natoms,ncg,icalc_type=-1,lreadaalibs=0;
  int lhybrid = ((simparms->icalc_type==5)||(simparms->icalc_type==6));
  double ttemp,tdt,tpext;
  double *px,*py,*pz,*vx,*vy,*vz,**eta,**veta,*bar,*vbar;
  FILE *fp;
#ifdef DEBUG
  double *amass = coords->amass;
#endif

  *estep   = 0;
  eta = coords -> eta;  veta = coords -> veta;
  bar = coords -> bar;  vbar = coords -> vbar;
  energies->deltae2 = 0.;

  ncg = simparms->ncg;
#ifdef DEBUG
  sprintf(line,"istart = %d",simparms->istart);
  md_stdout(line);
#endif
  if(simparms->istart == 1){
    sprintf(line,"Doing cold start from initial file \"%s\"",initfile);
    md_stdout(line);
    
    if((fp = fopen(initfile,"r")) == NULL){
      sprintf(line,"%s: can't open control file (%s)",command,initfile);
      md_error(line);
    }
    /* read header  */
    if(fgets(line,MAXLINELEN,fp) == NULL){
      sprintf(line,"can't read header from infile (%s)",initfile);
      md_error(line);
    }
    if((sscanf(line,"%d %d",&n,&ist)) != 2){
      sprintf(line,"Something wrong with header from infile (%s)",initfile);
      md_error(line);
    }

    if(lhybrid){
      if(n == simparms->naa){
	md_stdout("Reading atomistic positions from file and generating matching coarsegrained configuration");
        natoms=simparms->naa;
	coords = coords->next;
	lreadaalibs = 0;
      }else if(n == simparms->ncg){	
	md_stdout("Reading CG positions from file and AA positions from library files");
        natoms=simparms->ncg;
	lreadaalibs = 1;
      }else{
        sprintf(line,"Number of atoms in file %s (%d) does not match with number of CG atoms (%d) or AA atoms (%d) in topology files",
		initfile,n,simparms->ncg,simparms->naa);
        md_error(line);
      }
    }else{
      natoms=simparms->natoms;
      if(natoms != n){
        sprintf(line,"Different number of atoms natoms\n\tN in set file and parameter file = %d\n\tN in %s = %d",
		simparms->natoms,initfile,n);
        md_error(line);
      }
    }
    
    if(lreadaalibs){
      /* generate AA positions from library files of molecules */
      iat=0; icg=0; 
      coords=coords->next;
      for(i=0;i<(simparms->nspec/2);i++){
	px = coords->px+iat;  py = coords->py+iat; pz = coords->pz+iat;
	natoms=0;
	while( (simparms->ispecies[icg]==i) && (simparms->imolecule[icg]==0) ) natoms += simparms->napcg[icg++];
	readaapos(simparms->mole[i],px,py,pz,natoms);
	iat += natoms;
	while(simparms->imolecule[icg]!=0){
	  for(j=0;j<natoms;j++){
	    coords->px[iat] = px[j];
	    coords->py[iat] = py[j];
	    coords->pz[iat++] = pz[j];
	  }
	  imol = simparms->imolecule[icg];
	  while( (simparms->ispecies[icg]==i) && (imol == simparms->imolecule[icg]) ) icg++;
	}
      }
      for(i=0;i<coords->natoms*3;i++) coords->vx[i] = 0.0;  /*actually AA vel will be randomized */
      coords=coords->next;
      natoms=coords->natoms;
    }
    
    px = coords -> px;  py = coords -> py;  pz = coords -> pz;
    vx = coords -> vx;  vy = coords -> vy;  vz = coords -> vz;
    if(ist != simparms->istart){
      sprintf(line,"istart in initial file %d != %d in input file",
	      ist,simparms->istart);
      md_error(line);
    }
    
    for(i=0;i<natoms;i++){
      if(fgets(line,MAXLINELEN,fp) == NULL){
        sprintf(line,"Something is wrong with infile (%s) at line %d.",
		initfile,i+1);
        md_error(line);
      }
      
      if((sscanf(line,"%lg %lg %lg",&px[i],&py[i],&pz[i])) != 3){
        sprintf(line,"Something wrong with infile (%s) at line %d.",
		initfile,i+1);
        md_error(line);
      }
      vx[i] = vy[i] = vz[i] = 0.0;
    }
    if(lhybrid){
      if(lreadaalibs){
        coords=coords->next;
        check_aa_com(coords->next->px,coords->px,coords->amass,coords->next->amass,
            simparms->icglist,simparms->napcg,simparms->naa,simparms->ncg,0);
        coords=coords->next;
      }else{
	/* compute positions of the coarse grain sites */
        get_cgcom(simparms,coords,coords->next);
        coords=coords->next;
        for(i=0;i<ncg*3;i++) coords->vx[i] = 0.0;
      }
      /* compute angular momenta from the atomic velocities */
      getMitAA(simparms,coords->next,so3);
      getOmegaInit(simparms,coords->next,so3);
    }
#ifdef DEBUG
    for(i=0;i<simparms->natoms;++i){
      fprintf(stderr,"%4d%12.6lf%12.6lf%12.6lf%12.6lf\n",i,amass[i],px[i],py[i],pz[i]);
    }
#endif
    
#ifdef ORIG_HMAT
    for(i=0;i<3;i++){
      if(fgets(line,MAXLINELEN,fp) == NULL){
        sprintf(line,"Something is wrong with infile (%s) at line %d. (hmat)",
		initfile,i+simparms->natoms+1);
        md_error(line);
      }
      if((sscanf(line,"%lg %lg %lg",
		 &coords->hmat[i*3],&coords->hmat[i*3+1],&coords->hmat[i*3+2]))
         != 3){
        sprintf(line,"Something is wrong with infile (%s) at line %d. (hmat)",
		initfile,i+simparms->natoms+1);
        md_error(line);
      }
    }
#else
    for(i=0;i<9;i++){
      if((fscanf(fp,"%lg",&coords->hmat[i]))!=1){
        sprintf(line,"Something is wrong with infile (%s) at line %d. (hmat)",
		initfile,i+simparms->natoms+1);
        md_error(line);
      }
    }
    
#endif
    coords->pvol = log(get_deth(coords->hmat))/DIM;
    coords->vvol = 0.;
    gethinv9(coords->hmat,coords->hmati);
    for(i=0;i<simparms->nbar;i++) coords->bar[i] = coords->vbar[i] = 0.;
    if(lhybrid){
      for(i=0;i<simparms->nbar;i++) coords->next->bar[i] = coords->next->vbar[i] = 0.;
      for(i=0;i<9;++i){
	coords->next->hmat[i] = coords->hmat[i];
	coords->next->hmati[i] = coords->hmati[i];
      }
    }
    
    fclose(fp);
  } else {
    sprintf(line,"Doing warm start from restart file \"%s\"",initfile);
    md_stdout(line);
    /* read dump file */
    if((fp = fopen(initfile,"r"))==NULL){
      sprintf(line,"can't open restart file %s",initfile);
      md_error(line);
    }
    /* read header  */
    if(fgets(line,MAXLINELEN,fp) == NULL){
      sprintf(line,"can't read 1st line of header from infile (%s)",initfile);
      md_error(line);
    }
    if( sscanf(line ,"%d %d %d %d %d %lg %lg %lg",&simparms->istep,
	       &n,&nt,&nc,&nb,&ttemp,&tpext,&tdt) != 8){
      sprintf(line,"can't read in header file of %s (line 1)",initfile);
      md_error(line);
    }
    if(strstr(line,"HYBRID")!=NULL) icalc_type=5;
    if((icalc_type==5)&&(!lhybrid)){
      sprintf(line,"The restart file %s is from a hybrid MD run!",initfile);
      md_warning(line);
    }else if(icalc_type==-1 && lhybrid){
      sprintf(line,"The restart file %s is not from a hybrid MD run! Trying to read only AA part...",initfile);
      md_warning(line);
    }
    if((simparms->natoms != n)&&(simparms->naa != n)){
      sprintf(line,"Different number of atoms %d (initfile)",n);
      sprintf(line,"%s != %d (setfile+parmfile",line,simparms->natoms);
      md_error(line);
    }
    fgets(line,MAXLINELEN,fp);
    if(sscanf(line,"%lg %lg %lg %lg %lg %lg %lg %lg",
	      &(energies->avham),&(energies->avpotra),&(energies->avpoter),
	      &(energies->avzke),&(energies->avpotn),&(energies->avzken),
	      &(energies->avpotv),&(energies->avzkev)) != 8){
      sprintf(line,"while reading averages (line 2) (%s)",initfile);
      md_error(line);
    }
    fgets(line,MAXLINELEN,fp);
    if(sscanf(line,"%lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
	      &(energies->avpot_inter),&(energies->avpot_bond),
	      &(energies->avpot_bend),&(energies->avpot_tors),
	      &(energies->avpot_onfo),&(energies->avpot_onfo_e),
	      &(energies->avpot_elec),&(energies->avpot_recip),
	      &(energies->avpot_extern),&(energies->avpot_extern_e))!= 10){
      sprintf(line,"while reading averages (line 3) (missing extern?) (%s)",
	      initfile);
      md_error(line);
    }
    
    fgets(line,MAXLINELEN,fp);
    if(sscanf(line,"%lg %lg %lg %lg %lg",&(energies->avtemp),
	      &(energies->avprs),&(energies->avvol),
	      &(energies->econv),&(energies->econ2)) != 5){
      sprintf(line,"while reading averages (line 4) (%s)",initfile);
      md_error(line);
    }
    fgets(line,MAXLINELEN,fp);
    if(sscanf(line,"%lg %lg %lg %lg %lg %lg %lg %lg %lg",
	      &(energies->avprs_tensor[0]),&(energies->avprs_tensor[1]),
	      &(energies->avprs_tensor[2]),&(energies->avprs_tensor[3]),
	      &(energies->avprs_tensor[4]),&(energies->avprs_tensor[5]),
	      &(energies->avprs_tensor[6]),&(energies->avprs_tensor[7]),
	      &(energies->avprs_tensor[8])) != 9){
      sprintf(line,"while reading averages (line 5) (%s)\n",initfile);
      sprintf(line,"possible old restart file (add 9 elements after line 4)");
      md_error(line);
    }
    if(icalc_type==5 && lhybrid){
      energies=energies->next;
      fgets(line,MAXLINELEN,fp);
      if(sscanf(line,"%lg %lg %lg %lg %lg %lg %lg %lg",
		&(energies->avham),&(energies->avpotra),&(energies->avpoter),
	        &(energies->avzke),&(energies->avpotn),&(energies->avzken),
	        &(energies->avpotv),&(energies->avzkev)) != 8){
        sprintf(line,"while reading averages (line 6) (%s)",initfile);
        md_error(line);
      }
      fgets(line,MAXLINELEN,fp);
      if(sscanf(line,"%lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
	        &(energies->avpot_inter),&(energies->avpot_bond),
	        &(energies->avpot_bend),&(energies->avpot_tors),
	        &(energies->avpot_onfo),&(energies->avpot_onfo_e),
	        &(energies->avpot_elec),&(energies->avpot_recip),
	        &(energies->avpot_extern),&(energies->avpot_extern_e))!= 10){
        sprintf(line,"while reading averages (line 7) (missing extern?) (%s)",
	        initfile);
	md_error(line);
      }
      
      fgets(line,MAXLINELEN,fp);
      if(sscanf(line,"%lg %lg %lg %lg %lg",&(energies->avtemp),
	        &(energies->avprs),&(energies->avvol),
	        &(energies->econv),&(energies->econ2)) != 5){
        sprintf(line,"while reading averages (line 8) (%s)",initfile);
        md_error(line);
      }
      fgets(line,MAXLINELEN,fp);
      if(sscanf(line,"%lg %lg %lg %lg %lg %lg %lg %lg %lg",
	        &(energies->avprs_tensor[0]),&(energies->avprs_tensor[1]),
	        &(energies->avprs_tensor[2]),&(energies->avprs_tensor[3]),
	        &(energies->avprs_tensor[4]),&(energies->avprs_tensor[5]),
	        &(energies->avprs_tensor[6]),&(energies->avprs_tensor[7]),
	        &(energies->avprs_tensor[8])) != 9){
        sprintf(line,"while reading averages (line 9) (%s)",initfile);
        md_error(line);
      }
      fgets(line,MAXLINELEN,fp);
      if(sscanf(line,"%lg %lg %lg %lg %lg %lg %d",
	        &(energies->avzkecg),&(energies->avcgtra),
	        &(energies->avcgter),&(energies->vcgter),
	        &(energies->next->avcgter),&(energies->next->vcgter),&j) != 7){
        sprintf(line,"while reading hybrid CG inherent energies (line 10) (%s)\n possible old restart file (add 6 elements and a 1 after line 9)",initfile);
        md_error(line);
      }
      fgets(line,MAXLINELEN,fp);
      if(sscanf(line,"%lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
		&(energies->avso3_ham),&(energies->avso3_poter),&(energies->avso3_potra),&(energies->avso3_bond),
	        &(energies->avso3_bend),&(energies->avso3_tors),&(energies->avso3_onfo),&(energies->avso3_onfo_e),&(energies->avso3_ekin),
	        &(energies->avso3_temp)) != 10){
        sprintf(line,"while reading hybrid SO3 dynamics energies (line 11) (%s)\npossible old restart file (add 10 elements after line 10)",initfile);
        md_error(line);
      }
      energies=energies->next;
    }
    j=0;
    while(j<n){
      if(lhybrid && (n==simparms->naa)) coords = coords->next;
      px = coords -> px;  py = coords -> py;  pz = coords -> pz;
      vx = coords -> vx;  vy = coords -> vy;  vz = coords -> vz;
      for(i=0;i<coords->natoms;i++){
        fgets(line,MAXLINELEN,fp);
        if(sscanf(line,"%lg %lg %lg %lg %lg %lg"
		  ,&px[i],&py[i],&pz[i],&vx[i],&vy[i],&vz[i]) != 6){
          sprintf(line,"while reading in positions and velocites");
          sprintf(line,"%s at line %d (atom # %d)",line,j,j+5);
          md_error(line);
        }
	j++;
      }
      if(lhybrid) coords = coords->next;
    }
    if(lhybrid){
      if(icalc_type==5){  /* restart contains CG and AA pos+vel; now match AA c.o.m. to CG pos */
	check_aa_com(coords->px,coords->next->px,coords->next->amass,
		     coords->amass,simparms->icglist,simparms->napcg,simparms->naa,simparms->ncg,0);
	check_aa_com(coords->vx,coords->next->vx,coords->next->amass,
		     coords->amass,simparms->icglist,simparms->napcg,simparms->naa,simparms->ncg,1);
      }else{			   /* restart only contains AA pos+vel; now compute CG pos+vel  */
	get_cgcom(simparms,coords->next,coords);
	get_cgvelcom(simparms,coords->next,coords);
      }
    }
    
    /* read in hmatrix */
    for(i=0;i<3;i++){
      if(fgets(line,MAXLINELEN,fp) == NULL){
        sprintf(line,"something is wrong with infile \"%s\" at line %d (hmat)",
		initfile,i+simparms->natoms+5);
        md_error(line);
      }
      if(sscanf(line,"%lg %lg %lg",
		&coords->hmat[i*3],&coords->hmat[i*3+1],&coords->hmat[i*3+2])
         !=3){
        sprintf(line,"something is wrong with infile (%s) at line %d (hmat)",
		initfile,i+simparms->natoms+5);
        md_error(line);
      }
    }
    gethinv9(coords->hmat,coords->hmati);
    if(lhybrid){
      for(i=0;i<9;++i){
	coords->next->hmat[i] = coords->hmat[i];
	coords->next->hmati[i] = coords->hmati[i];
      }
    }
    
    /* read in vol of hmat */
    for(i=0;i<3;i++){
      if(fgets(line,MAXLINELEN,fp) == NULL){
        sprintf(line,"something is wrong in file \"%s\" at line %d(vvol9)",
		initfile,i+simparms->natoms+5+3);
        md_error(line);
      }
      if(sscanf(line,"%lg %lg %lg",&coords->vvol9[i*3],&coords->vvol9[i*3+1],
		&coords->vvol9[i*3+2]) !=3){
        sprintf(line,"something is wrong in file \"%s\" at line %d (vvol9)\n",
		initfile,i+simparms->natoms+5+3);
        sprintf(line,
		"%s possible old restart file (add nine elements after hmat)",
		line);
        md_error(line);
      }
    }
    
    if(fabs((ttemp-simparms->temp)/TZ(ttemp))>ERRMAX){
      sprintf(line,"input temp %g != %g stored temp",simparms->temp,ttemp);
      md_warning(line);
      if(simparms->istart==2){
	if(lhybrid){
	  samvel(simparms,coords->next,0);
	  get_cgvel(coords,simparms->icglist,simparms->napcg,coords->natoms,coords->iselect);
	}else{
	  samvel(simparms,coords,0);
	}
      }
    }
    if(fabs((tpext-simparms->pext)/TZ(tpext))>(ERRMAX*10000)){
      sprintf(line,"input pressure %g != %g stored pressure",
	      simparms->pext,tpext);
      md_warning(line);
    }
    if(fabs((tdt-simparms->dt)/TZ(tdt))>ERRMAX){
      sprintf(line,"old time step %g != %g new time step",tdt,simparms->dt);
      md_warning(line);
    }
    if(simparms->nchain != nc ||  nt != simparms->ntherm ){
      if( (icalc_type==-1 && lhybrid) && (nt == simparms->ntherm-coords->ntherm) ){
        sprintf(line,"Only thermostats for the atomistic representation are read from restart\n\t ");
        sprintf(line,"thermostats for coarsegrained representation are randomized\n\t ");
        md_warning(line);
        samveta(simparms,coords);
        for(j=coords->ntherm;j<nt+coords->ntherm;j++){
          for(i=0;i<nc;i++){
            fscanf(fp,"%lg %lg",&eta[j][i],&veta[j][i]);
          }
        }
      }else{
        sprintf(line,"Different number of thermostats (will randomize all)\n\t ");
        sprintf(line,"%s nchain = %d != %d (being used)\n\t",line,nc,
		simparms->nchain);
        sprintf(line,"%s ntherm = %d != %d (being used)\n\t",line,nt,
		simparms->ntherm);
        md_warning(line);
	/*      for(j=0;j<simparms->ntherm;j++)	*/  
	/* randomize instead of setting to zero */
	/*        for(i=0;i<simparms->nchain;i++) eta[j][i] = veta[j][i] = 0.; */  /* changed by Bernd  */
        samveta(simparms,coords);
        for(j=0;j<nt;j++) for(i=0;i<nc;i++) fscanf(fp,"%*g %*g");
      }
    } else { 
      for(j=0;j<nt;j++){
        for(i=0;i<nc;i++){
          fscanf(fp,"%lg %lg",&eta[j][i],&veta[j][i]);
        }
      }
    }
    
#ifdef ISO
    fscanf(fp,"%lg %lg",&coords->pvol,&coords->vvol);
    
    if(fabs(coords->pvol-log(get_deth(coords->hmat))/DIM)
       /get_deth(coords->hmat)>ERRMAX){
      sprintf(line,"Cell size don't match %12g (from cell coords) != %12g",
	      log(get_deth(coords->hmat))/DIM,coords->pvol);
      sprintf(line,"%s\n\t Resetting Volume to %g",
	      line,get_deth(coords->hmat));
      md_warning(line);
      coords->pvol = log(get_deth(coords->hmat))/DIM;
      coords->vvol = 0.;
    }
#endif
    if(simparms->nbar != nb){
      sprintf(line,"Different number of barostats\n\t");
      sprintf(line,"%s nbar = %d != %d (being used)",line,nb,simparms->nbar);
      md_warning(line);
    }
    
    /* zero new barostats */
    for(i=nb;i<simparms->nbar;i++) bar[i] = vbar[i] = 0.;
    /* read all available barostats */
    nb = MIN(simparms->nbar,nb);
    for(i=0;i<nb;i++) fscanf(fp,"%lg %lg",&bar[i],&vbar[i]);
    
    /* read SO3 dynamics angular velocities for hybrid MD */
    if(lhybrid){
      if(icalc_type==5){
        for(i=0;i<coords->natoms;i++){
          if(fscanf(fp,"%lg %lg %lg",&(so3->omega[i][0]),&(so3->omega[i][1]),&(so3->omega[i][2])) != 3){
            sprintf(line,"while reading in SO3 angular velocities");
            md_error(line);
          }
        }
      }else{
        /* compute angular momenta from the atomic velocities */
        getMitAA(simparms,coords,so3);
        getOmegaInit(simparms,coords->next,so3);
      }
    }
    
    fclose(fp);
    keep(initfile);
  }
  
#ifdef DEBUG
  /* NOTE: DEBUG WORKS NOT FOR HYBRID BECAUSE px runs out of bounds (naa+ncg; not n) */
  sprintf(line,"%d %p %p %p",simparms->natoms,px,py,pz);
  md_stdout(line);
  for(i=0;i<n;i++){
    sprintf(line,"%i %g %g %g %g %g %g",i,px[i],py[i],pz[i],vx[i],vy[i],vz[i]);
    md_stdout(line);
  }
  sprintf(line,"cell coords %g %g %g\t %g %g %g\t %g %g %g",
	  coords->hmat[0],coords->hmat[1],coords->hmat[2],
	  coords->hmat[3],coords->hmat[4],coords->hmat[5],
	  coords->hmat[6],coords->hmat[7],coords->hmat[8]);
  md_stdout(line);
  sprintf(line,"cellinv coords %g %g %g\t %g %g %g\t %g %g %g",
	  coords->hmati[0],coords->hmati[1],coords->hmati[2],
	  coords->hmati[3],coords->hmati[4],coords->hmati[5],
	  coords->hmati[6],coords->hmati[7],coords->hmati[8]);
  md_stdout(line);

  sprintf(line,"thermostats = %d by %d",simparms->ntherm,simparms->nchain);
  md_stdout(line);
  for(i=0;i<simparms->ntherm;i++){
    for(j=0;j<simparms->nchain;j++){
      sprintf(line,"%d %d %g %g",i,j,eta[i][j],veta[i][j]);
      md_stdout(line);
    }
  }
  sprintf(line,"barostats = %d",simparms->nbar); md_stdout(line);
  for(i=0;i<simparms->nbar;i++){
    sprintf(line,"%d %g %g",i,bar[i],vbar[i]); md_stdout(line);
  }
#endif
  
  sprintf(line,"%d Atoms read in with a Cell Volume of %g",
	  simparms->natoms,get_deth(coords->hmat));
  md_stdout(line);
}
/*------------------------------------------------------------------*/
